import unittest
import numpy as np
import numpy.testing as npt
from wilson.run.smeft import beta
from wilson.util import smeftutil
from wilson.util.common import EFTutil
from wilson.run.smeft.tests import test_beta
from wilson.test_wilson import get_random_wc
from wilson import wcxf
from numbers import Number

C = test_beta.C.copy()
for i in C:
    if i in smeftutil.WC_keys_2f + smeftutil.WC_keys_4f:
        # make Wilson coefficients involving fermions complex!
        C[i] = C[i] + 1j*C[i]

class TestSymm(unittest.TestCase):
    def test_keys(self):
        # check no parameter or WC was forgotten in the C_symm_keys lists
        self.assertEqual(
          set(smeftutil.C_keys),
          {c for cs in smeftutil.C_symm_keys.values() for c in cs}
        )

    def test_symmetrize_symmetric(self):
        a = np.array([[1, 2, 3], [2, 4, 5], [3, 5, 6]])
        npt.assert_array_equal(EFTutil.symmetrize_2(a), a)
        b = np.array([[1, 2, 3], [0, 4, 5], [0, 0, 6]])
        npt.assert_array_equal(EFTutil.symmetrize_2(b), a)

    def test_symmetrize_hermitian(self):
        a = np.array([[1, 2j, 3j], [-2j, 4, 5j], [-3j, -5j, 6]])
        npt.assert_array_equal(EFTutil.symmetrize_2(a), a)
        b = np.array([[1, 2j, 3j], [0, 4, 5j], [0, 0, 6]])
        npt.assert_array_equal(EFTutil.symmetrize_2(b), a)

    def test_symmetrize_C(self):
        C_symm = smeftutil.symmetrize(C)
        # check all keys are present
        self.assertSetEqual(set(C.keys()), set(C_symm.keys()))
        for i, v in C_symm.items():
            # check trivial cases are the same
            if i in smeftutil.C_symm_keys[0] + smeftutil.C_symm_keys[1] + smeftutil.C_symm_keys[3]:
                if smeftutil.C_keys_shape[i] == 1:
                    self.assertEqual(v, C[i])
                else:
                    npt.assert_array_equal(v, C[i])
            # check symmetric
            elif i in smeftutil.C_symm_keys[9]:
                npt.assert_array_equal(v, v.T)
            # check hermitian
            elif i in smeftutil.C_symm_keys[2]:
                npt.assert_array_equal(v, v.T.conj())
            # check 2 identical FFbar
            elif i in smeftutil.C_symm_keys[4]:
                npt.assert_array_equal(v, v.transpose((2, 3, 0, 1)))
                npt.assert_array_equal(v, v.transpose((1, 0, 3, 2)).conj())
            # check 2 independent FFbar
            elif i in smeftutil.C_symm_keys[5]:
                npt.assert_array_equal(v, v.transpose((1, 0, 3, 2)).conj())
            # check special case ee
            elif i in smeftutil.C_symm_keys[6]:
                npt.assert_array_equal(v, v.transpose((2, 3, 0, 1)))
                npt.assert_array_equal(v, v.transpose((0, 3, 2, 1)))
                npt.assert_array_equal(v, v.transpose((2, 1, 0, 3)))
            # check special case qque
            elif i in smeftutil.C_symm_keys[7]:
                npt.assert_array_equal(v, v.transpose((1, 0, 2, 3)))
            # check special case qqql
            elif i in smeftutil.C_symm_keys[8]:
                # see eq. (10) of arXiv:1405.0486
                npt.assert_array_almost_equal(v + v.transpose((1, 0, 2, 3)), v.transpose((1, 2, 0, 3)) + v.transpose((2, 1, 0, 3)), decimal=15)

    def test_wcxf2array(self):
        wc = get_random_wc('SMEFT', 'Warsaw', 160)
        C = smeftutil.wcxf2arrays_symmetrized(wc.dict)
        d = smeftutil.arrays2wcxf_nonred(C)
        self.assertEqual(wc.dict.keys(), d.keys())
        for k, v in wc.dict.items():
            self.assertAlmostEqual(v, d[k], msg=f"Failed for {k}")

    def test_wcxf2array_incomplete(self):
        wc = wcxf.WC('SMEFT', 'Warsaw', 160, {'G': 1e-10})
        C = smeftutil.wcxf2arrays_symmetrized(wc.dict)
        d = smeftutil.arrays2wcxf_nonred(C)
        self.assertEqual(wc.dict.keys(), d.keys())
        for k, v in d.items():
            self.assertEqual(v, wc[k], msg=f"Failed for {k}")
            self.assertIsInstance(v, Number)


class TestKeysShapesSymm(unittest.TestCase):

    # names of SM parameters
    SM_keys = ['g', 'gp', 'gs', 'Lambda', 'm2', 'Gu', 'Gd', 'Ge',]

    # names of WCs with 0, 2, or 4 fermions (i.e. scalars, 3x3 matrices,
    # and 3x3x3x3 tensors)
    WC_keys_0f = ["G", "Gtilde", "W", "Wtilde", "phi", "phiBox", "phiD", "phiG",
                  "phiB", "phiW", "phiWB", "phiGtilde", "phiBtilde", "phiWtilde",
                  "phiWtildeB"]
    WC_keys_2f = ["uphi", "dphi", "ephi", "eW", "eB", "uG", "uW", "uB", "dG", "dW",
                  "dB", "phil1", "phil3", "phie", "phiq1", "phiq3", "phiu", "phid",
                  "phiud", "llphiphi"]
    WC_keys_4f = ["ll", "qq1", "qq3", "lq1", "lq3", "ee", "uu", "dd", "eu", "ed",
                  "ud1", "ud8", "le", "lu", "ld",  "qe", "qu1", "qd1", "qu8",
                  "qd8", "ledq", "quqd1", "quqd8", "lequ1", "lequ3", "duql",
                  "qque", "qqql", "duue"]

    C_keys = SM_keys + WC_keys_0f + WC_keys_2f + WC_keys_4f
    WC_keys = WC_keys_0f + WC_keys_2f + WC_keys_4f

    C_keys_shape = {
       'g': 1,
       'gp': 1,
       'gs': 1,
       'Lambda': 1,
       'm2': 1,
       'Gu': (3, 3),
       'Gd': (3, 3),
       'Ge': (3, 3),
       'G': 1,
       'Gtilde': 1,
       'W': 1,
       'Wtilde': 1,
       'phi': 1,
       'phiBox': 1,
       'phiD': 1,
       'phiG': 1,
       'phiB': 1,
       'phiW': 1,
       'phiWB': 1,
       'phiGtilde': 1,
       'phiBtilde': 1,
       'phiWtilde': 1,
       'phiWtildeB': 1,
       'uphi': (3, 3),
       'dphi': (3, 3),
       'ephi': (3, 3),
       'eW': (3, 3),
       'eB': (3, 3),
       'uG': (3, 3),
       'uW': (3, 3),
       'uB': (3, 3),
       'dG': (3, 3),
       'dW': (3, 3),
       'dB': (3, 3),
       'phil1': (3, 3),
       'phil3': (3, 3),
       'phie': (3, 3),
       'phiq1': (3, 3),
       'phiq3': (3, 3),
       'phiu': (3, 3),
       'phid': (3, 3),
       'phiud': (3, 3),
       'llphiphi': (3, 3),
       'll': (3, 3, 3, 3),
       'qq1': (3, 3, 3, 3),
       'qq3': (3, 3, 3, 3),
       'lq1': (3, 3, 3, 3),
       'lq3': (3, 3, 3, 3),
       'ee': (3, 3, 3, 3),
       'uu': (3, 3, 3, 3),
       'dd': (3, 3, 3, 3),
       'eu': (3, 3, 3, 3),
       'ed': (3, 3, 3, 3),
       'ud1': (3, 3, 3, 3),
       'ud8': (3, 3, 3, 3),
       'le': (3, 3, 3, 3),
       'lu': (3, 3, 3, 3),
       'ld': (3, 3, 3, 3),
       'qe': (3, 3, 3, 3),
       'qu1': (3, 3, 3, 3),
       'qd1': (3, 3, 3, 3),
       'qu8': (3, 3, 3, 3),
       'qd8': (3, 3, 3, 3),
       'ledq': (3, 3, 3, 3),
       'quqd1': (3, 3, 3, 3),
       'quqd8': (3, 3, 3, 3),
       'lequ1': (3, 3, 3, 3),
       'lequ3': (3, 3, 3, 3),
       'duql': (3, 3, 3, 3),
       'qque': (3, 3, 3, 3),
       'qqql': (3, 3, 3, 3),
       'duue': (3, 3, 3, 3),
    }

    # names of Wilson coefficients with the same fermionic symmetry properties
    C_symm_keys = {}
    # 0 0F scalar object
    C_symm_keys[0] = WC_keys_0f + ['g', 'gp', 'gs', 'Lambda', 'm2',]
    # 1 2F general 3x3 matrix
    C_symm_keys[1] = ["uphi", "dphi", "ephi", "eW", "eB", "uG", "uW", "uB", "dG", "dW", "dB", "phiud"] + ['Gu', 'Gd', 'Ge']
    # 2 2F Hermitian matrix
    C_symm_keys[2] = ["phil1", "phil3", "phie", "phiq1", "phiq3", "phiu", "phid",]
    # 3 4F general 3x3x3x3 object
    C_symm_keys[3] = ["ledq", "quqd1", "quqd8", "lequ1", "lequ3", "duql", "duue"]
    # 4 4F two identical ffbar currents
    C_symm_keys[4] = ["ll", "qq1", "qq3", "uu", "dd",]
    # 5 4F two independent ffbar currents
    C_symm_keys[5] = ["lq1", "lq3", "eu", "ed", "ud1", "ud8", "le", "lu", "ld", "qe", "qu1", "qd1", "qu8", "qd8",]
    # 6 4F two identical ffbar currents - special case Cee
    C_symm_keys[6] = ["ee",]
    # 7 4F Baryon-number-violating - special case Cqque
    C_symm_keys[7] = ["qque",]
    # 8 4F Baryon-number-violating - special case Cqqql
    C_symm_keys[8] = ["qqql",]
    # 9 2F symmetric matrix
    C_symm_keys[9] = ["llphiphi"]

    def test_keys(self):
        self.assertEqual(self.SM_keys, smeftutil.dim4_keys)
        self.assertEqual(self.WC_keys_0f, smeftutil.WC_keys_0f)
        self.assertEqual(self.WC_keys_2f, smeftutil.WC_keys_2f)
        self.assertEqual(self.WC_keys_4f, smeftutil.WC_keys_4f)
        self.assertEqual(self.C_keys, smeftutil.C_keys)
        self.assertEqual(self.WC_keys, smeftutil.WC_keys)

    def test_shape(self):
        self.assertEqual(self.C_keys_shape, smeftutil.C_keys_shape)

    def test_symm_keys(self):
        self.assertEqual(self.C_symm_keys.keys(), smeftutil.C_symm_keys.keys())
        for k, v in self.C_symm_keys.items():
            self.assertEqual(set(v), set(smeftutil.C_symm_keys[k]))

    def test_needs_padding(self):
        self.assertEqual(smeftutil._needs_padding, False)



"""Utility modules for wilson."""

from wilson.util.common import EFTutil as EFTutil

smeftutil = EFTutil(
    'SMEFT',
    'Warsaw',
    dim4_keys_shape = {
       'g': 1,
       'gp': 1,
       'gs': 1,
       'Lambda': 1,
       'm2': 1,
       'Gu': (3, 3),
       'Gd': (3, 3),
       'Ge': (3, 3),
    },
    dim4_symm_keys =  {
        0: ['g', 'gp', 'gs', 'Lambda', 'm2'],
        1: ['Gu', 'Gd', 'Ge'],
    },
)
wetutil = EFTutil(
    'WET',
    'JMS',
    dim4_keys_shape = {
        'e': 1,
        'gs': 1,
        'Md': (3, 3),
        'Mu': (2, 2),
        'Me': (3, 3),
        'Mnu': (3, 3),
    },
    dim4_symm_keys = {
        0: ['e', 'gs'],
        1: ['Md', 'Mu', 'Me', 'Mnu'],
    },
)



"""Utility functions and dictionaries useful for the manipulation of SMEFT Wilson coefficients.
"""

import numpy as np
import wilson


def flavor_rotation(C_in, Uq, Uu, Ud, Ul, Ue):
    """Gauge-invariant $U(3)^5$ flavor rotation of all Wilson coefficients."""
    C = {}
    # nothing to do for purely bosonic operators
    for k in wilson.util.smeftutil.WC_keys_0f:
        if k in C_in:
            C[k] = C_in[k]
    # see 1704.03888 table 4 (but staying SU(2) invariant here)
    # LR
    for k in ['ephi', 'eW', 'eB']:
        if k in C_in:
            C[k] = Ul.conj().T @ C_in[k] @ Ue
    for k in ['uphi', 'uW', 'uB', 'uG']:
        if k in C_in:
            C[k] = Uq.conj().T @ C_in[k] @ Uu
    for k in ['dphi', 'dW', 'dB', 'dG']:
        if k in C_in:
            C[k] = Uq.conj().T @ C_in[k] @ Ud
    # LL
    for k in ['phil1', 'phil3']:
        if k in C_in:
            C[k] = Ul.conj().T @ C_in[k] @ Ul
    for k in ['phiq1', 'phiq3']:
        if k in C_in:
            C[k] = Uq.conj().T @ C_in[k] @ Uq
    if 'llphiphi' in C_in:
        C['llphiphi'] = Ul.T @ C_in['llphiphi'] @ Ul
    # RR
    if 'phie' in C_in:
        C['phie'] = Ue.conj().T @ C_in['phie'] @ Ue
    if 'phiu' in C_in:
        C['phiu'] = Uu.conj().T @ C_in['phiu'] @ Uu
    if 'phid' in C_in:
        C['phid'] = Ud.conj().T @ C_in['phid'] @ Ud
    if 'phiud' in C_in:
        C['phiud'] = Uu.conj().T @ C_in['phiud'] @ Ud
    # 4-fermion
    if 'll' in C_in:
        C['ll'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ul, Ul, Ul.conj(), Ul.conj(), C_in['ll'])
    if 'ee' in C_in:
        C['ee'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ue, Ue, Ue.conj(), Ue.conj(), C_in['ee'])
    if 'le' in C_in:
        C['le'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ul, Ue, Ul.conj(), Ue.conj(), C_in['le'])
    if 'qq1' in C_in:
        C['qq1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Uq, Uq.conj(), Uq.conj(), C_in['qq1'])
    if 'qq3' in C_in:
        C['qq3'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Uq, Uq.conj(), Uq.conj(), C_in['qq3'])
    if 'dd' in C_in:
        C['dd'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ud, Ud, Ud.conj(), Ud.conj(), C_in['dd'])
    if 'uu' in C_in:
        C['uu'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Uu, Uu.conj(), Uu.conj(), C_in['uu'])
    if 'ud8' in C_in:
        C['ud8'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Ud, Uu.conj(), Ud.conj(), C_in['ud8'])
    if 'ud1' in C_in:
        C['ud1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Ud, Uu.conj(), Ud.conj(), C_in['ud1'])
    if 'qu1' in C_in:
        C['qu1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Uu, Uq.conj(), Uu.conj(), C_in['qu1'])
    if 'qu8' in C_in:
        C['qu8'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Uu, Uq.conj(), Uu.conj(), C_in['qu8'])
    if 'qd1' in C_in:
        C['qd1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Ud, Uq.conj(), Ud.conj(), C_in['qd1'])
    if 'qd8' in C_in:
        C['qd8'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Ud, Uq.conj(), Ud.conj(), C_in['qd8'])
    if 'quqd1' in C_in:
        C['quqd1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Ud, Uq.conj(), Uq.conj(), C_in['quqd1'])
    if 'quqd8' in C_in:
        C['quqd8'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Ud, Uq.conj(), Uq.conj(), C_in['quqd8'])
    if 'lq1' in C_in:
        C['lq1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ul, Uq, Ul.conj(), Uq.conj(), C_in['lq1'])
    if 'lq3' in C_in:
        C['lq3'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ul, Uq, Ul.conj(), Uq.conj(), C_in['lq3'])
    if 'ld' in C_in:
        C['ld'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ul, Ud, Ul.conj(), Ud.conj(), C_in['ld'])
    if 'lu' in C_in:
        C['lu'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ul, Uu, Ul.conj(), Uu.conj(), C_in['lu'])
    if 'qe' in C_in:
        C['qe'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Ue, Uq.conj(), Ue.conj(), C_in['qe'])
    if 'ed' in C_in:
        C['ed'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ue, Ud, Ue.conj(), Ud.conj(), C_in['ed'])
    if 'eu' in C_in:
        C['eu'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ue, Uu, Ue.conj(), Uu.conj(), C_in['eu'])
    if 'ledq' in C_in:
        C['ledq'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ue, Uq, Ul.conj(), Ud.conj(), C_in['ledq'])
    if 'lequ1' in C_in:
        C['lequ1'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ue, Uu, Ul.conj(), Uq.conj(), C_in['lequ1'])
    if 'lequ3' in C_in:
        C['lequ3'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Ue, Uu, Ul.conj(), Uq.conj(), C_in['lequ3'])
    if 'duql' in C_in:
        C['duql'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Ul, Ud, Uq, C_in['duql'])
    if 'qque' in C_in:
        C['qque'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Ue, Uq, Uu, C_in['qque'])
    if 'qqql' in C_in:
        C['qqql'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uq, Ul, Uq, Uq, C_in['qqql'])
    if 'duue' in C_in:
        C['duue'] = np.einsum('jb,ld,ia,kc,ijkl->abcd', Uu, Ue, Ud, Uu, C_in['duue'])
    return C



import unittest
from wilson.util import wet_jms, wetutil
import numpy as np
from wilson.test_wilson import get_random_wc
import numbers
from numpy.testing import assert_array_equal

class TestWETutil(unittest.TestCase):
    def test_scalar2array(self):
        d = {'bla_123': 3, 'blo': 5j}
        da = wet_jms._scalar2array(d)
        self.assertEqual(da['blo'], 5j)
        self.assertIn('bla', da)
        self.assertEqual(da['bla'].shape, (3, 3, 3))
        self.assertTrue(np.isnan(da['bla'][0, 0, 0]))
        self.assertEqual(da['bla'][0, 1, 2], 3)

    def test_padding(self):
        wc = get_random_wc('WET', 'JMS', 90)
        C = wetutil.wcxf2arrays(wc.dict)
        C_shapes = {
            k: 1
            if isinstance(v, numbers.Number)
            else v.shape
            for k,v in C.items()
        }
        self.assertGreaterEqual(wetutil.C_keys_shape.items(), C_shapes.items())
        C_padded = wetutil.pad_C(C)
        C_padded_shapes = {
            k: 1
            if isinstance(v, numbers.Number)
            else v.shape
            for k,v in C_padded.items()
        }
        self.assertEqual(
            3, min((min(v) for v in C_padded_shapes.values() if v != 1))
        )
        for k, v in wetutil.unpad_C(C_padded).items():
            assert_array_equal(C[k], v)

    def test_symmetrize(self):
        wc = get_random_wc('WET', 'JMS', 90)
        C = wetutil.wcxf2arrays_symmetrized(wc.dict)
        d = wetutil.arrays2wcxf_nonred(C)
        self.assertEqual(wc.dict, d)


class TestKeysShapesSymm(unittest.TestCase):

    # names of Wilson coefficients with the same fermionic symmetry properties
    # numbering is inspired by the corresponding categorization in SMEFT
    C_symm_keys = {}
    # 0 0F scalar object
    C_symm_keys[0] = ["G", "Gtilde"] + ["e", "gs"]
    # 1 2F general 3x3 matrix
    C_symm_keys[1] = ["egamma", "uG","dG", "ugamma", "dgamma"] + ["Mnu", "Mu",
        "Me", "Md"]
    # 3 4F general 3x3x3x3 object
    C_symm_keys[3] = ['S1udRR', 'S1udduRR', 'S8udRR', 'S8udduRR', 'SedRL',
    'SedRR', 'SeuRL', 'SeuRR', 'SnueduRL', 'SnueduRR', 'TedRR', 'TeuRR',
    'TnueduRR', 'V1udduLR', 'V8udduLR', 'VnueduLL', 'VnueduLR',
    'SuddLL', 'SduuLL', 'SduuLR', 'SduuRL', 'SdudRL', 'SduuRR',]
    # 4 4F two identical ffbar currents
    # hermitian currents
    C_symm_keys[4] = ['VuuRR', 'VddRR', 'VuuLL', 'VddLL']
    # non-hermitian currents
    C_symm_keys[41] = ['S1ddRR', 'S1uuRR', 'S8uuRR', 'SeeRR', 'S8ddRR']
    # 5 4F two independent ffbar currents, hermitian
    C_symm_keys[5] = ["VnueLL", "VnuuLL", "VnudLL", "VeuLL", "VedLL", "V1udLL",
            "V8udLL", "VeuRR", "VedRR", "V1udRR", "V8udRR", "VnueLR",
            "VeeLR", "VnuuLR", "VnudLR", "VeuLR", "VedLR", "VueLR", "VdeLR",
            "V1uuLR", "V8uuLR", "V1udLR", "V8udLR", "V1duLR", "V8duLR",
            "V1ddLR", "V8ddLR"]
    # 6 4F two identical ffbar currents + Fierz symmetry
    C_symm_keys[6] = ['VeeLL', 'VeeRR', 'VnunuLL']
    # 4F antisymmetric in first 2 indices
    C_symm_keys[71] = ['SuudLR', 'SuudRL', 'SdduRL']

    def test_symm_keys(self):
        self.assertEqual(self.C_symm_keys.keys(), wetutil.C_symm_keys.keys())
        for k, v in self.C_symm_keys.items():
            self.assertEqual(set(v), set(wetutil.C_symm_keys[k]))

    def test_needs_padding(self):
        self.assertEqual(wetutil._needs_padding, True)



import unittest
from wilson import wcxf

class TestBases(unittest.TestCase):
    def test_bases(self):
        for basis in wcxf.Basis.instances.values():
            try:
                basis.validate()
            except:
                self.fail(f"Basis {basis.eft}-{basis.basis} failed to validate")
